#include "actuator.h"
/**
 Update: 
	tamvu - Sept 27 2011
	added: 
		returning k hashed IPs and k index of the prefixes on BGP table 
		returning all k latency for a query 
 */

Actuator::Actuator(const char *topo_Filename, const char *pref_Filename){

	//Actuator(); 
		//Do nothing
	initMASK(); //CONSTANT MATRIX
	number_of_dropped_insert =0; 
	number_of_successful_insert =0; 
	number_of_dropped_query =0; 
	number_of_successful_query =0; 
	act_time = 0; 
	as_v.clear();		//flush as vector 
	//Read prefix from file, initialize time 
	curr_prefix.init(pref_Filename,act_time); 
	cout <<"Number of Pref: "<<	curr_prefix.entryList.size()<<endl;
	//Read topology, Compute Routing table, 	
	topology.init(topo_Filename); 
	//Create a list of AS from PrefixTable and Topology file
	cout <<"Start init ASes" << endl; 
	initASes(); 
	
}
/****************************************************/
/*
Initialize AS_Vector with ASes's ID and their prefixes
*/
void Actuator::initASes(){
	//Walk through TOPOLOGY to get all AS with its sequence 
	cout <<"Walk through TOPOLOGY to get all AS with its sequence" << endl; 
	for(u32b i=0; i<topology.nas; i++){
		//Creat an AS with ID from topology object
		AutoSystem tmpAS(topology.idx_as_arr[i]); 
		//add it into AS vector of the actuator 
		as_v.push_back(tmpAS); 
	}
	//Walk through PREFIX to assign it into coressponding AS
	cout << "Walk through PREFIX to assign it into coressponding AS" <<endl; 
	for(u32b i=0; i<curr_prefix.entryList.size();i++){
		//cout <<"At prefix number:  "<< i <<endl; 
		Prefix_entry tmpPrefEntry; 					//tempt prefix 
		tmpPrefEntry = curr_prefix.entryList[i];	
		//Put that prefix into the corresponding AS
		asIdx tmpASidx = topology.cvertASnum2idx(tmpPrefEntry.orgin_AS);

		if (tmpASidx >= topology.nas){  //That AS ID doesn't not found in the TOPOLOGY, which means ERROR
			cerr << "**********************ERROR*****************************************"<<endl;
			cerr << "AS NUMBER: " << tmpPrefEntry.orgin_AS << "DOESN'T EXIST ON TOPOLOGY "<<endl; 
			cerr << "********************************************************************"<<endl;
			return; 
		}
		//Found ASIndex -> Insert that Prefix in to the AS
		Cidr cidr_2bInsert = {tmpPrefEntry.prefix, tmpPrefEntry.mask_bit}; 
		as_v[tmpASidx].addPrefix(cidr_2bInsert); 
	}
	cout << "done init*********************************" <<endl; 
}	
/****************************************************/
/*
	
	Going through event file line by line and executing corresponding event
	Input file format: 
		<OpCode, GUID, SourceASNum, TimeStamp>
		Opcode: I,Q,U,R  corresponding to Insert, Query, Update, RefressPrefixTable
	Note: Refresh Prefix table is not implemented

*/
void Actuator::exec(const char *event_Filename){

	//Readfile
	ifstream event_FHdlr(event_Filename);
	string tempt_line; //iterator to walk through event file 
	//Command parameters
	char optCode; 
	u64b inpGUID, inpTimeStamp;
	asNum sourceASNum; 
	//counter for debuging purpose
	u64b processedEvent =0; 
	
	
	//output File
	string queFN(event_Filename);
	queFN += "_queryLatency.txt"; 
	ofstream queryLatFile(queFN.c_str()); 
	cout <<" Openning Event file !!!" <<endl;
	//ifstream event_FHdlr(event_Filename);
	if (!queryLatFile.is_open()){
		cerr << "Can't open QUERY OUTPUT File !!! Returning..." <<endl;
		return; 
	}
	
	if (event_FHdlr.is_open()){
		while (event_FHdlr.good()){
			tempt_line.clear(); //init
			getline(event_FHdlr,tempt_line); //read from file
			//parse the command	
			vector<string> strTemp_v; 
			//std::toupper((unsigned char)c)
			Common::str2StrArr(tempt_line,' ', strTemp_v); 
			optCode = toupper((unsigned char)(strTemp_v[0][0]));  //optCode
			inpGUID = atol(strTemp_v[1].c_str()); 
			sourceASNum = (asNum)atoi(strTemp_v[2].c_str());
			//cout << " Parsed successfully " << endl;
			inpTimeStamp = atol(strTemp_v[3].c_str()); 
			processedEvent++;
			if (processedEvent % 5000 ==0)
				cout << "Processed: " << processedEvent << endl; 
			switch (optCode){
				case 'I':
					//cout<< " Inserting GUID: " << inpGUID << endl; 
					insertGUID_fromAS(inpGUID,sourceASNum, inpTimeStamp); 
					break;
				case 'U':
					updateGUID_fromAS(inpGUID,sourceASNum, inpTimeStamp); 
					break;
				case 'Q':
					if (DEBUG ==1) cout << sourceASNum <<endl;
					//cout <<tempt_line <<endl;  
					//execute the query and return latency to the file 
					queryLatFile << queryGUID_fromAS(inpGUID,sourceASNum, inpTimeStamp) << endl;
					break;
				case 'R':
					//Donothing now
					break;
				default:
					cout << "Some thing wrong with optCode !!!!!"; 
			}
		}
		//cout << "out of while loop" << endl;
	}
	else{
		cerr << "Can't open EVENT File !!! Returning..." <<endl;
		return; 
	}
	//cout << "Get query Latency " << endl; 
	queryLatFile.close(); 
	getGUIDperASdistribution(event_Filename);
	
}
/****************************************************/
/* 
    Insert GUID with sourceASnum info.
	Replicate <GUID> to K_NUM places 
	
	update Oct 5, 2011: 
		handle ip hole problem after MAX_TRY_NUM
*/
void Actuator::insertGUID_fromAS(u64b GUID, asNum sourceASnum, u64b time){

	for (u8b i=0; i<K_NUM; i++){
		//get K hash of GUID (hash(GUID) 
		u32b destIP = Hash::hashG2IP(GUID, i); 
		if (DEBUG ==1) cout << "destIP : " <<destIP <<endl;
		u8b tryNum = 0;
		u8b longestPref = 0; 
		asNum destAS = 0; 
		Cidr longestCidrPref; 
		//loop to find destination AS
		while ((tryNum <MAX_TRY_NUM) && (longestPref == 0 )) {
			//cout << "GUID: " << GUID << ", try  Num: "<< (unsigned int) tryNum << ", destIP" << (unsigned int)destIP << endl; 
			longestPrefixMatching(destIP, longestPref, destAS, longestCidrPref); 
		//	cout << "GUID: " << GUID << ", try  Num: "<< (unsigned int) tryNum << ", destIP" << (unsigned int)destIP << ", dest AS: "<< destAS << endl;
		//	if (DEBUG ==1) cout << "longestPref:  " << longestPref <<endl;
			if (longestPref == 0){	//no prefix is found 
				destIP = Hash::hashIP2IP(destIP); 
				tryNum++; 	
			}
			
		}		//end loop 
		
		if (longestPref == 0) {   //no AS is found after MAX_TRY
			//find the closest prefix
			u32b minDistance = 4294967295U; 
			asNum asNumber = 0; 
			//Walk through all prefixes PREFIX to find the closest prefix 
			for(u32b i=0; i<curr_prefix.entryList.size();i++){
				if ( minDistance > Common::ipDistanceIPtoPrefix(curr_prefix.entryList[i],destIP) ){
					minDistance = Common::ipDistanceIPtoPrefix(curr_prefix.entryList[i],destIP); 
					asNumber =  curr_prefix.entryList[i].orgin_AS; 
					longestCidrPref.prefix =  curr_prefix.entryList[i].prefix; 
					longestCidrPref.mask_bit =  curr_prefix.entryList[i].mask_bit; 
				}
			}
			cout << "used ipDistance !!! Distance is : "<< (unsigned int) minDistance << ", AS Number:  " << (unsigned int)asNumber << endl; 
			destAS = asNumber; 
		}
		
		if (destAS == 0) {
			cout << " Can't find destination AS to insert GUID : "<<  GUID << endl; 
			cout << " Exiting now from insertGUID_fromAS !!!! " << endl; 
			abort(); 
		}			
		else 
			insertGUID_fromAS_toAS(GUID, sourceASnum, destAS, longestCidrPref, time); 
	}
	
}
/****************************************************/
/*
	Get called after destination ASNum is found 
	It computes the path between the two ASes and deliver the insert/update
	
	Return 0 if success, and 1 otherwise
	Note: NO ACK is exectuted. 
		  This only called when prefix is on that destination AS 
	
*/
int Actuator::insertGUID_fromAS_toAS(u64b GUID, asNum sourceASnum, asNum destASnum, Cidr longestCidrPref, u64b original_time){
	//if (GUID == 611) cout << "611 is inserted to : " <<destASnum <<endl;
	asIdx srAsIdx, dstAsIdx;
	u64b tmp_time = original_time; 
	srAsIdx = topology.cvertASnum2idx(sourceASnum);
	dstAsIdx = topology.cvertASnum2idx(destASnum); 
	if (DEBUG ==1) cout << "found dest AS: " << destASnum << endl;
	//if there is no path  between the two ASnums
	if (topology.getShortestDistance(sourceASnum,destASnum) == MAX_EDGE_WEIGHT){
		number_of_dropped_insert++;
		if (DEBUG ==1) {
			cerr << "**********************PATH NOT FOUND ERROR**************************"<<endl;
			cerr << "NO PATH FROM AS# " << sourceASnum << " TO AS# "<< destASnum << endl; 
			cerr << "********************************************************************"<<endl;
		}
		return 1;
	}
	//get the path from source to dest 
	vector<asIdx> pathBetweenASes= topology.getASidxOnPathfromAs2AS(sourceASnum, destASnum); 
	vector<asIdx>::iterator asIdx_iter; 
	//update intermediate ASes along the path
 
	 if (DEBUG ==1) topology.printPath(pathBetweenASes); 

	for (asIdx_iter = pathBetweenASes.begin(); asIdx_iter <pathBetweenASes.end(); asIdx_iter++){
		//for each AS 
		if (((*asIdx_iter) != srAsIdx) && ((*asIdx_iter) != dstAsIdx)){
			//update packet counter 
			as_v[*asIdx_iter].packet_counter++; 
			//update AS's clock if needed 
			as_v[*asIdx_iter].update_local_clock(original_time + topology.getShortestDistUsingASidx(srAsIdx,*(asIdx_iter))); 
		}
	}
	//insert GUI to destination AS 
	u32b inASPrefixIndex; 
	//get the prefix index 
	as_v[dstAsIdx].prefix_lookup(longestCidrPref, inASPrefixIndex); 
	//now insert GUID
	as_v[dstAsIdx].add_GUID_entry(GUID, inASPrefixIndex, original_time+ topology.getShortestDistance(sourceASnum,destASnum));
	//increase number of inserted	
	number_of_successful_insert++; 
	return 1; 
}
/****************************************************/
/*
	
	Do longest prefix matching for an IP 
	return: Maxlenght (which can be ZERO) 
			and the coresponding asNumber

*/

void Actuator::longestPrefixMatching(u32b IP, u8b &maxPrefLen, asNum &asNumber, Cidr &maxPrefCidr){
	maxPrefLen =0; 
	asNumber =0; 
	//Walk through all prefixes PREFIX to find the longest one 
	for(u32b i=0; i<curr_prefix.entryList.size();i++){
		u32b bitMask = curr_prefix.entryList[i].mask_bit-1;
		u32b left = curr_prefix.entryList[i].prefix & MASK[bitMask];
		u32b right = IP & MASK[bitMask]; 
		if ((left == right) && (maxPrefLen < bitMask+1)){
			maxPrefLen = bitMask + 1; 
			asNumber = curr_prefix.entryList[i].orgin_AS; 
			maxPrefCidr.mask_bit = curr_prefix.entryList[i].mask_bit;
			maxPrefCidr.prefix = curr_prefix.entryList[i].prefix; 
		}
	}
}

/*
	Simulate a querying process from source AS
	Input: GUID, sourceAS and query time
	Output: Query latency  OR ZERO if no result was found 
	Note: Query timeStamp must be smaller than GUID's Update time 
*/
string Actuator::queryGUID_fromAS(u64b GUID, asNum sourceASnum, u64b timeStamp){
	string result = intToString(GUID); // in the form of : GUID destAS1 matched_CirdIP1 bitMask1 latency1 destAS2 matched_CirdIP2 bitMask2 latency2 ... destASK matched_CirdIPK bitMaskK latencyk minimumLatency
	vector<asNum> canditate_AS_v; 
	//Updated 09/28/2011: this vector is to store longest matched CIDR
	vector<Cidr> candidate_AS_longest_matched_prefix; 
	
	u8b num_of_can =0; //number of candatate AS
	vector<asNum>::iterator can_AS_it; 
	//Get K destination IP from GUID by hashing 
	for (u8b i=0; i<K_NUM; i++){
		//get K hash of GUID (hash(GUID) 
		u32b destIP = Hash::hashG2IP(GUID, i); 
		u8b tryNum = 0;
		u8b longestPref = 0; 
		asNum destAS = 0; 
		Cidr longestCidrPref; 
		//loop to find destination AS that has longest matched prefix  
		while ((tryNum <MAX_TRY_NUM) && (longestPref == 0 )) {
			//cout << " tryNum: " << (unsigned int)tryNum << endl; 
			longestPrefixMatching(destIP, longestPref, destAS, longestCidrPref); 
			if (longestPref == 0){	//no prefix is found 
				destIP = Hash::hashIP2IP(destIP); 
				tryNum++; 	
			}
		}		//end loop 
					
		if (longestPref == 0) {   //no AS is found after MAX_TRY
			//find the closest prefix
			u32b minDistance = 4294967295U; 
			asNum asNumber = 0; 
			//Walk through all prefixes PREFIX to find the closest prefix 
			for(u32b i=0; i<curr_prefix.entryList.size();i++){
				if ( minDistance > Common::ipDistanceIPtoPrefix(curr_prefix.entryList[i],destIP) ){
					minDistance = Common::ipDistanceIPtoPrefix(curr_prefix.entryList[i],destIP); 
					asNumber =  curr_prefix.entryList[i].orgin_AS; 
					longestCidrPref.prefix =  curr_prefix.entryList[i].prefix; 
					longestCidrPref.mask_bit =  curr_prefix.entryList[i].mask_bit; 
				}
			}
			destAS = asNumber; 
		}		
		if (destAS == 0) {
			cout << " Can't find destination AS to insert GUID : "<<  GUID << endl; 
			cout << " Exiting now from queryGUID_fromAS !!!! " << endl; 
			abort(); 
		}	
		
		canditate_AS_v.push_back(destAS); //put the destination ASnumber in to the vector
		//updated 09/28/2011
		candidate_AS_longest_matched_prefix.push_back(longestCidrPref); 
		
	}
	// Updated 09/28/2011: to this point we have the vector of AS candidates that can answer that query 
	
	//get the one that has shortest distance/latency 
	u64b minLatency = MAX_EDGE_WEIGHT; 
	asNum minLatecyASNum; 
	
	//cout << "candidate vector size : " << canditate_AS_v.size() << endl; 
	
	for (u8b i = 0; i<canditate_AS_v.size(); i++){
		//cout << "Inside For loop " <<endl; 
		u64b tmpLatency = topology.getShortestDistance(sourceASnum, canditate_AS_v[i]);
		
		
		//update 10/28/2011: output hopcount 		
		vector<asIdx> shortestPath = topology.getASidxOnPathfromAs2AS(sourceASnum, canditate_AS_v[i]);		
		
		
		//cout << "tmp Latency :" << (unsigned long) tmpLatency <<endl; 
		//updated 09/28/2011: concatinate to result 
		//update 10/28/2011:  output hopcount 	
		{
		// print that vector out  
			//Appending to result: ASnumb || longestPrefixMatch.prefix || longestPrefixMatch.mask || latencyBetweenThe2AS
								//        || hopcount || list all hops
								
		result = result + "\t" + intToString(canditate_AS_v[i]) + "\t" + intToString(candidate_AS_longest_matched_prefix[i].prefix) 
				 + "\t" + intToString((u32b)candidate_AS_longest_matched_prefix[i].mask_bit) + "\t" + intToString((u64b)tmpLatency) 
				 + "\t" + intToString((u32b)shortestPath.size()) + "\t" + topology.printPath(shortestPath); 
		 
		}
		//cout << "result  :" << result <<endl; 
		if (minLatency > tmpLatency){
			minLatecyASNum = canditate_AS_v[i]; 
			minLatency = tmpLatency; 
		}
	}
	//cout << "Check if it is acceptable distance " <<endl;
	if (minLatency == MAX_EDGE_WEIGHT) {  //No AS was found
		number_of_dropped_query++; 
		if (DEBUG == 1) {
			cerr << "************ CAN'T execute QUERY Operation **************" <<endl; 
			cerr << "No AS was found for GUID : " << GUID << endl; 
			cerr << "*********************************************************" <<endl; 
		}
		return 0; 
	}
	// An AS with minimum latency was found 
	//Go to the dest AS and get the GUID
	//if (GUID == 611) cout <<"Querry found dest AS for GUID 611 is : "<< minLatecyASNum << endl; 
	u64b minimumQueryLatency = queryGUID_fromAS_toAS(GUID, sourceASnum, minLatecyASNum, timeStamp);	
	result = result + "\t" + intToString((u64b)minimumQueryLatency);
	////cout << result <<endl; 
	return result; 
}
/****************************************************/
/*
	AS to AS update for the query 
*/
u64b Actuator::queryGUID_fromAS_toAS(u64b GUID, asNum sourceASnum, asNum destASnum, u64b timeStamp){
	//cout << "in QueryGUID_fromAS_toAS " << endl;
	//if (GUID == 611) cout << "Going to AS number " << destASnum << "for querr 611" <<endl; 

	asIdx srAsIdx, dstAsIdx;
	u64b tmp_time = timeStamp; 
	srAsIdx = topology.cvertASnum2idx(sourceASnum);
	dstAsIdx = topology.cvertASnum2idx(destASnum); 
	u64b minLatency = topology.getShortestDistance(sourceASnum,destASnum);

	//get the path from source to dest 
	vector<asIdx> pathBetweenASes= topology.getASidxOnPathfromAs2AS(sourceASnum, destASnum); 
	vector<asIdx>::iterator asIdx_iter; 
	//update intermediate ASes along the path 
	for (asIdx_iter = pathBetweenASes.begin(); asIdx_iter <pathBetweenASes.end(); asIdx_iter++){
		if (((*asIdx_iter) != srAsIdx) && ((*asIdx_iter) != dstAsIdx)){  //avoid source and destASes 
			//update packet counter passed through the AS 
			as_v[*asIdx_iter].packet_counter++; 
			//update AS's clock if needed 
			as_v[*asIdx_iter].update_local_clock(timeStamp + topology.getShortestDistUsingASidx(srAsIdx,*(asIdx_iter))); 
		}
	}
	//Lookup on destination AG 
	u32b inASPrefixIndex; 
	u32b inASinPrefixGUIDindex; 
	//get the prefix index 
	//cout << "GUID lookup on that AS" <<endl; 
	bool found = as_v[dstAsIdx].blind_GUID_lookup(GUID, inASPrefixIndex, inASinPrefixGUIDindex); 
	//GUID not found 
	if (!found) {
			number_of_dropped_query++;
			cout <<"not found" <<endl;  	
			if (DEBUG == 1) {
			cerr << "************ CAN'T execute QUERY Operation **************" <<endl; 
			cerr << "GUID NOT FOUND WITHIN an AS  " << GUID << endl; 
			cerr << "*********************************************************" <<endl; 
			}
			return 0; 	
	}
	//checking GUID timeStamp
	if (found) {
		//check timeStamp
		
		u64b guidUpdateTime = as_v[dstAsIdx].as_pref_vt[inASPrefixIndex].GUID_list[inASinPrefixGUIDindex].updated_time; 
		if (timeStamp + minLatency < guidUpdateTime){ //ask before GUID avaiable
			number_of_dropped_query++; 	
			if (DEBUG == 1) {
			cerr << "************ CAN'T execute QUERY Operation **************" <<endl; 
			cerr << "GUID comes after the it is queried. GUID: " << GUID << endl; 
			cerr << " Query time: " << timeStamp << "vs. GUID update time: " <<guidUpdateTime << endl; 
			cerr << "*********************************************************" <<endl; 
			}
		//	cout << "not found -- DROPPED due to timestamp!!! <GUID: "<< GUID << " ,sourceAS: "<<sourceASnum<<" ,destAS: " << destASnum<< " ,minLatency: " << minLatency <<" >" <<endl; 
			return 0; 	
			
		}
		//cout << "found !!! <GUID: "<< GUID << " ,sourceAS: "<<sourceASnum<<" ,destAS: " << destASnum<< " ,minLatency: " << minLatency <<" >" <<endl; 
	}
	//now GUID is indeed found 
	as_v[dstAsIdx].update_local_clock(timeStamp + minLatency); 
	//increase number of inserted	
	number_of_successful_query++; 
	return minLatency;
}
/*
	Update GUID from AS
	Simply call InsertGUID_fromAS
*/
void Actuator::updateGUID_fromAS(u64b GUID, asNum sourceASnum, u64b time){
	insertGUID_fromAS(GUID,sourceASnum,time);
}

/****************************************************/
/*
initialize the MASK array
(UTILITY FUNCTION)
*/
void Actuator::initMASK(){
	u32b tmp =0;	 
	for (u8b i=0; i<32; i++){
		u32b tmp2 = (1); 
			tmp2 = tmp2 << (31-i) ; 
		tmp = tmp2 | tmp;
		if (DEBUG ==1) cout << tmp <<endl; 
		MASK[i]=tmp; 
	}
}


void Actuator::getGUIDperASdistribution(const char *event_Filename){
	cout <<" In    getGUIDperASdistribution !!!!"<<endl;
	string GpAdistFileName(event_Filename);
	GpAdistFileName+="_perAS.txt";

	ofstream outFile(GpAdistFileName.c_str()); 
	if (!outFile.is_open()){
		cerr << "Can't open perAS File for OUTPUT !!! Returning..." <<endl;
		return; 
	}
		
	//for each AS
	u64b  guidCount;
	for(u32b i = 0; i<as_v.size(); i++){
		//cout << "at AS: " <<i <<endl; 
		//count number of GUID
		guidCount =0;
		for (u32b j=0; j<as_v[i].as_pref_vt.size(); j++)
			guidCount += as_v[i].as_pref_vt[j].GUID_list.size(); 
		// print to file, in the format of:
		// Total number of GUID  [space] number of packet 
		outFile << as_v[i].asID << " "<<guidCount <<" "<< as_v[i].packet_counter << endl; 
	}
	outFile.close();
	
}

/**
Utility function to convert an u64b number to string 
*/

string Actuator::intToString(u64b num){
	std::string s;
	std::stringstream out;
	out << num;
	s = out.str();
	return s; 
}

string Actuator::intToString(u32b num){
	std::string s;
	std::stringstream out;
	out << num;
	s = out.str();
	return s; 
}
int main(int argc, char *argv[]){
	//Prefix_entry tmp; 
        //tmp.prefix=2783313920;
	//tmp.mask_bit=16; 
	//cout << "distance : " << Common::ipDistanceIPtoPrefix(tmp,2783248384) << endl; 	
	//return 0; 

	char *pFilename; //event file name 
	if (argc == 0) {
		cout << "Event file name is missing !!!! " <<endl;
		return 1; 
	}
	else {
			pFilename = argv[1]; //get event file name
			cout << "Input event file : " <<pFilename << endl; 
			ifstream my_event_FHdlr(pFilename);
				if (!my_event_FHdlr.is_open()){
					cerr << "Can't open QUERY OUTPUT File !!! Returning..." <<endl;
				return 0; 
				}
			my_event_FHdlr.close();
	}
			
	Actuator actor("new1_Oct04_topology.data", "prefix_2_1_11_adjusted.data"); 
	actor.exec(pFilename);
	cout << "FINISHED !!!!!!!!"<<endl; 

	return 0; 
}
