/*
	Actuator reads Topology, prefixTable and Event list from Files and then act accordingly 
*/
#ifndef ACTUATOR_H
#define ACTUATOR_H

#include "common.h"
#include "hash.h"
#include "announced_prefix_table.h"
#include "topology.h"
#include "as.h"
#include <sstream>

using namespace std;

u32b MASK[32]; 

class Actuator
{
public:
	//c/dstructors
	//Actuator(); 
	Actuator(const char *topo_Filename, const char *pref_Filename);	

	//*********************OPERATIONs*******************
	void exec(const char *event_Filename);
	void stop();
	void havest_latency(); 
	void havest_traffic(); 

	//*********************MEMBERs*******************
	PrefixTable curr_prefix; 
	Topology topology;
	vector<AutoSystem> as_v; 
	u64b act_time; 

	u64b number_of_dropped_insert; 
	u64b number_of_successful_insert; 

	u64b number_of_dropped_query;
	u64b number_of_successful_query;
	
	

	 

	
	void initASes(); 
	
	void insertGUID_fromAS(u64b GUID, asNum sourceASnum, u64b time);
	int insertGUID_fromAS_toAS(u64b GUID, asNum sourceASnum, asNum destASnum, Cidr longestCidrPref, u64b time); 
	
	string queryGUID_fromAS(u64b GUID, asNum sourceASnum, u64b timeStamp);  //return required time
	u64b queryGUID_fromAS_toAS(u64b GUID, asNum sourceASnum, asNum destASnum, u64b timeStamp); 

	void updateGUID_fromAS(u64b GUID, asNum sourceASnum, u64b time);
	
	void refrestPrefixTable(); 

	//Harvesting
	void getGUIDperASdistribution(const char *event_Filename);


	//utilities function 
	void longestPrefixMatching(u32b IP, u8b &maxPrefLen, asNum &asNumber, Cidr &returnedCidr);
	void initMASK(); 
	string intToString(u64b num); 
	string intToString(u32b num); 
	

};
	
	
#endif //ACTUATOR_H